#include <Arduino.h>
#include "OpenTherm.h"
#include <ESP8266WiFi.h>
#include <PubSubClient.h>
#include <DHT.h>

#define DHTPIN 12          // GPIO12 ESP8266 D6
#define DHTTYPE DHT11     // typ czujnika 
DHT dht(DHTPIN, DHTTYPE); // definicja czujnika
// this part schall be changed ----------------------------------
IPAddress local_IP(x, x, x, x); //do uzupłęnienia
IPAddress gateway(x, x, x, x); //do uzupłęnienia
IPAddress subnet(x, x, x, x); //do uzupłęnienia
IPAddress mqtt_server(x, x, x, x); //do uzupłęnienia
const char* ssid = "xxx"; //do uzupłęnienia
const char* password = "xxx"; //do uzupłęnienia
const char* mqtt_client_id = "xxx"; // Unique client id
const int mqtt_Port = xxx; //do uzupłęnienia
const char* mqtt_User = "xxx"; //do uzupłęnienia
const char* mqtt_Password = "xxx"; //do uzupłęnienia
//float maxVent = 2.49; //it means 249 m/h3 - max avilable flow in my Brink - not used/needed
const unsigned long readPeriod = 7000; // 1000 = every second; set between 1000 - 5000
const unsigned long readPeriod_bypass = 120000;// Set +15000 - OT disconnection needed for bypass work
const unsigned long readPeriodDHT = 150000;

//-----------------------------------------------------------------------
const char* mqtt_topic_in = "brink/+/set"; //subscribe commands from Openhab

unsigned long startTime;
unsigned long currentTime;
unsigned long startTimeDHT;

const int inPin = 4; //ESP8266 D2
const int outPin = 5; //ESP8266 D1
OpenTherm ot(inPin, outPin);
WiFiClient espClient;
PubSubClient mqttClient(espClient);


float tIn, tIn_old = 0; // temp external
float tOut, tOut_old = 0; //temp internal
bool fault, fault_old = 1;
bool vmode, vmode_old = 0;
bool sbypass, sbypass_old = 1;
bool pbypass, pbypass_old;
bool filter, filter_old = 1;
int output, output_old = 0;
int input, input_old = 0;
int gvent, gvent_old = 0;
int pressin, pressin_old = 0;
int pressout, pressout_old = 0; // uint8_t
int vstep1, vstep1_old = 0;
int vstep2, vstep2_old = 0;
int vstep3, vstep3_old = 0;
int vstep8, vstep8_old = 1;
float vstep9, vstep9_old = 0;
int byPassStatus, byPassStatus_old = 0;
int tU4, tU4_old = 1;
float tU5, tU5_old = 1;
int cvol, cvol_old = 0;
int RPMin, RPMin_old;
int RPMout, RPMout_old;
int fcode, fcode_old = 0;
int msg, msg_old = 0;
int param1, param1_old;
long lRssi, lRssi_old = 0;
int cpod_old, cpod = 0;
float t, t_old = 0;
int h, h_old = 0;
//boolean retained = false;
const int del = 20;
long timeOn;
int datas;

void ICACHE_RAM_ATTR handleInterrupt() {
  ot.handleInterrupt();
}


void MqttReconnect() {
  while (!mqttClient.connected()) {
    Serial.print("Connecting to MQTT Broker: ");
    Serial.println( mqtt_server );
    //delay(500);
    if (mqttClient.connect(mqtt_client_id, mqtt_User, mqtt_Password)) {
      Serial.print("CONNECTED, topic: ");
      mqttClient.subscribe(mqtt_topic_in);
      Serial.println( mqtt_topic_in );
      Serial.println(mqttClient.state());
    } else {
      Serial.print("failed: ");
      Serial.print(mqttClient.state());
      Serial.println(" try again...");
      delay(5000);
    }
  }
  Serial.println(" ok...");
}



void MqttCallback(char* topic, byte* payload, unsigned int length) {
  payload[length] = 0;

  // Changing selected Brink Renovent HR parameters requested by OpenHab
  if (strcmp(topic, "brink/VentNomValue/set") == 0)  ot.setVentilation( atoi((char *)payload) ); // uint8_t
  if (strcmp(topic, "brink/U1/set") == 0) ot.setBrinkTSP(U1, atoi((char *)payload) );
  if (strcmp(topic, "brink/U2/set") == 0) ot.setBrinkTSP(U2, atoi((char *)payload) );
  if (strcmp(topic, "brink/U3/set") == 0) ot.setBrinkTSP(U3, atoi((char *)payload) );
  if (strcmp(topic, "brink/I8/set") == 0) ot.setBrinkTSP(I8, atoi((char *)payload) );  // bypass
  if (strcmp(topic, "brink/U4/set") == 0) ot.setBrinkTSP(U4, atoi((char *)payload) * 2 );

 if (strcmp(topic, "brink/U5/set") == 0) {
    ot.setBrinkTSP(U5, atoi((char *)payload) * 2 );
    refreshAll(); // change of U5 triggers refresh of other paramteres
  }


  if (strcmp(topic, "brink/REBOOT/set") == 0)
  {
    if ( *payload == '1' )
    {
      Serial.println( "-=REBOOT=-" );
      ESP.restart();
    }
  }
  /* if (strcmp(topic, "brink/SCAN/set") == 0)
    {
    if ( *payload == '1' )
    {
      Serial.println( "-=SCAN=-" );
      for (int i = 0; i < 90; i++)
        {
        scanner_TSP(i);
        scan(i);
        delay(20);
        }
    }
    }*/
}

void setup()
{
  delay(2500);
  Serial.begin(115200);
  Serial.println("Start");
  WiFi.mode(WIFI_STA);
  WiFi.begin(ssid, password);
  if (WiFi.config(local_IP, gateway, subnet))
  {
    Serial.println("Static IP Configured");
  }
  else {
    Serial.println("Static IP Configuration FAILED");
  }
  while (WiFi.status() != WL_CONNECTED)
  {
    Serial.print(".");
    delay(500);
  }
  Serial.println("");
  Serial.print("IP Addresss: ");
  Serial.println(WiFi.localIP());
  Serial.print("MAC Addresss: ");
  Serial.println(WiFi.macAddress());
  mqttClient.setServer(mqtt_server, mqtt_Port);
  mqttClient.setKeepAlive(30);
  mqttClient.setBufferSize(1024);
  mqttClient.setCallback(MqttCallback);
  Serial.println(mqttClient.state());
  MqttReconnect();

  dht.begin();
  ReadDHT();
  startTime = millis();
  ot.begin(handleInterrupt);

  // Floor heting pump switch - option
  //pinMode(HWCPin, OUTPUT);
  //digitalWrite(HWCPin, HIGH);
  timeOn = ((millis() / 1000) / 60);
  mqttClient.publish("brink/TimeON/get", String(timeOn).c_str());

  Serial.print("Number of available VentTSP ID's: ");
  Serial.println((int)ot.getTSPNumber()); // number of available VentTSP ID's


    //for (int i = U1; i != TempPostHeater; i++)
    //for (int i = 0; i < 73; i++)
    //{
     //scanner_TSP(i);
    //Serial.print("2TSP ID: ");
    //Serial.print(static_cast<BrinkTSPindex>(i));
    //Serial.print(" ==> DATA: ");
    //Serial.println(ot.getBrink2TSP(static_cast<BrinkTSPindex>(i)));
      //delay(200);
    //}
    /*for (int i = 0; i < 130; i++)
    {
      scan(i);
      delay(5);
    }
  */
}

void loop()
{
  double loopTime = millis();
  if (!mqttClient.connected())
  {
    MqttReconnect();
    mqttClient.publish("brink/OpenThermStatus/get", "WORK");

  }
  mqttClient.loop();
  //ReadBrinkParameters();

  currentTime = millis();

  if (currentTime - startTimeDHT >= readPeriodDHT)
  {
    ReadDHT();
    startTimeDHT = currentTime;
    timeOn = ((millis() / 1000) / 60);
    mqttClient.publish("brink/TimeON/get", String(timeOn).c_str());
  }

  if (currentTime - startTime >= readPeriod)
  {
    ReadBrinkParameters();
    Monitor();
    lRssi = WiFi.RSSI(); //long
    if ( abs(lRssi - lRssi_old) > 2) //reduce data publication due frequent slight changes of signal
    {
      mqttClient.publish("brink/Wifi/get", String(lRssi).c_str());
      Serial.println( lRssi );
      lRssi_old = lRssi;
    }
    startTime = currentTime;
  }

  loopTime = (millis() - loopTime) / 1000;
  if (loopTime > 1)
  {
    Serial.print( "Loop took: " );
    Serial.print( loopTime );
    Serial.println(" seconds");
  }
}

// refresh slow changing parameters on request by every change of U5
void refreshAll()
{

  mqttClient.publish("brink/TempSupplyIn/get", String(tIn).c_str());

  mqttClient.publish("brink/TempExhIn/get", String(tOut).c_str());

  mqttClient.publish("brink/FauiltIndication/get", String(fault).c_str());

  mqttClient.publish("brink/VentilationMode/get", String(vmode).c_str());

  mqttClient.publish("brink/BypassStaus/get", String(sbypass).c_str());

  mqttClient.publish("brink/FilterDirty/get", String(filter).c_str());

  mqttClient.publish("brink/VentilationNominalValue/get", String(gvent * 4.5).c_str());

  mqttClient.publish("brink/CurrentVolume/get", String(cvol).c_str());

  mqttClient.publish("brink/FaultCode/get", String(fcode).c_str());

  mqttClient.publish("brink/OperationMsg/get", String(msg).c_str());

  mqttClient.publish("brink/I1/get", String(param1 - 100).c_str());

  mqttClient.publish("brink/U1/get", String(vstep1).c_str());

  mqttClient.publish("brink/U2/get", String(vstep2).c_str());

  mqttClient.publish("brink/U3/get", String(vstep3).c_str());

  mqttClient.publish("brink/U4/get", String(tU4 / 2).c_str());

  mqttClient.publish("brink/U5/get", String(tU5 / 2).c_str());

  mqttClient.publish("brink/Wifi/get", String(lRssi).c_str());

  mqttClient.publish("brink/OpenThermStatus/get", "WORK");
}

void ReadDHT()
{
  t  = dht.readTemperature();
  h  = dht.readHumidity();

  if ( abs(t) != abs(t_old) ) {
    mqttClient.publish("ESP/Temp/get", String(t).c_str());
    t_old = t;
  }

  if ( abs(h) != abs(h_old)) {
    mqttClient.publish("ESP/Humid/get", String(h).c_str());
    h_old = h;
  }

  if (isnan(h) || isnan(t)) {
    Serial.println("NOT WORKING!");
  }
  else {
    Serial.print("Temp: ");
    Serial.print(t);
    Serial.print(" ^C ");
    Serial.print("Wilg: ");
    Serial.print(h);
    Serial.println(" % ");
  }
}


//Reading selected Brink Renovent HR parameters, Mqtt publication only if a value has changed
void ReadBrinkParameters()
{
  delay(del);
  tIn  = ot.getVentSupplyInTemperature();
  if ( abs(tIn) != abs(tIn_old)) {  //reduce data publication due frequent slight changes of temp
    mqttClient.publish("brink/TempSupplyIn/get", String(tIn).c_str());
    tIn_old = tIn;
  }



  delay(del);
  tOut = ot.getVentExhaustInTemperature();
  if ( abs(tOut) != abs(tOut_old)) {   //reduce data publication due frequent slight changes of temp
    mqttClient.publish("brink/TempExhIn/get", String(tOut).c_str()); // String.toCharArray()
    tOut_old = tOut;
  }

  //delay(del);
  //vstep8 = ot.getVentExhaustOutTemperature();
  //mqttClient.publish("brink/TempExhOUT/get", String(vstep8).c_str());

  //delay(del);
  //fault = ot.getFaultIndication();
  //if (fault != fault_old) {
  //mqttClient.publish("brink/FauiltIndication/get", String(fault).c_str());
  //fault_old = fault;
  //}

  /*  vmode = ot.getVentilationMode();
    if (vmode != vmode_old) {
      mqttClient.publish("brink/VentilationMode/get", String(vmode).c_str());
      vmode_old = vmode;
    }

  */
  //delay(del);
  //sbypass = ot.getBypassAutomaticStatus();         // ot.getBypassStatus() - this method does not work ot.getBrinkTSP(BypassStatus)
  //if (sbypass != sbypass_old) {
  //mqttClient.publish("brink/BypassStaus/get", String(sbypass).c_str());
  //sbypass_old = sbypass;
  // }

  //delay(del);
  //pbypass = ot.getBypassPosition();
  //if (pbypass != pbypass_old) {
  //mqttClient.publish("brink/BypassPosition/get", String(pbypass).c_str());
  //pbypass_old = pbypass;
  //}


  delay(del);
  filter = ot.getDiagnosticIndication();
  if (filter != filter_old) {
    mqttClient.publish("brink/FilterDirty/get", String(filter).c_str());
    filter_old = filter;
  }

  delay(del);
  gvent =  ot.getVentilation();
  if (gvent != gvent_old) {
    mqttClient.publish("brink/VentilationNominalValue/get", String(gvent * 4.5).c_str());
    gvent_old = gvent;
  }

  //delay(del);
  //pressin = ot.getBrinkTSP(CPID);
  // if ( abs(pressin - pressin_old) > 0 ) { //reduce data publication due frequent slight changes of preassure
  //mqttClient.publish("brink/CPID/get", String(pressin).c_str());
  // pressin_old = pressin;
  // }
  //delay(del);
  //pressout = ot.getBrinkTSP(CPOD);
  // if ( abs(pressout - pressout_old) > 0)  {  //reduce data publication due frequent slight changes of preassure
  //mqttClient.publish("brink/CPOD/get", String(pressout).c_str());
  //    pressout_old = pressout;
  // }


  delay(del);
  cvol = ot.getBrinkTSP(CurrentVol);
  if (cvol != cvol_old) {
    mqttClient.publish("brink/CurrentVolume/get", String(cvol).c_str());
    cvol_old = cvol;
  }
 // delay(del);
 // cpod = scanner_TSP((int)48);
 // if (cpod != cpod_old) {
  //  mqttClient.publish("brink/CPOD/get", String(cpod).c_str());
   // cpod_old = cpod;
 //}
  //delay(del);
  //fcode = ot.getVentFaultCode();
  //if (fcode != fcode_old) {
  //mqttClient.publish("brink/FaultCode/get", String(fcode).c_str());
  //fcode_old = fcode;
  //}
  delay(del);
  vstep1 = ot.getBrinkTSP(U1);
  if (vstep1 > 0) {
    mqttClient.publish("brink/U1/get", String(vstep1).c_str());
    vstep1_old = vstep1;

  }
  delay(del);
  vstep2 = ot.getBrinkTSP(U2);
  if (vstep2 > 0) {
    mqttClient.publish("brink/U2/get", String(vstep2).c_str());
    vstep2_old = vstep2;

  }
  delay(del);
  vstep3 = ot.getBrinkTSP(U3);
  if (vstep3 != vstep3_old) {
    mqttClient.publish("brink/U3/get", String(vstep3).c_str());
    vstep3_old = vstep3;

  }
  delay(del);
  tU4 = ot.getBrinkTSP(U4);
  if (tU4 != tU4_old) {
    mqttClient.publish("brink/U4/get", String(tU4 / 2).c_str());
    tU4_old = tU4;

  }
  delay(del);
  tU5 = ot.getBrinkTSP(U5);
  if (tU5 != tU5_old) {
    mqttClient.publish("brink/U5/get", String(tU5 / 2).c_str());
    tU5_old = tU5;
  }

  delay(del);
  param1 = ot.getBrinkTSP(I5);
  // if (param1 != param1_old) {
  mqttClient.publish("brink/I5/get", String(param1).c_str());
  //   param1_old = param1;
  // }
  delay(del);
  output = ot.getBrinkTSP(CurrentOutputVol);
  if (output != output_old) {
    mqttClient.publish("brink/CurrentOutputVol/get", String(output).c_str());
    output_old = output;
  }
  delay(del);
  input = ot.getBrinkTSP(CurrentInputVol2);
  if (input != input_old) {
    mqttClient.publish("brink/CurrentInputVol/get", String(input).c_str());
    input_old = input;
  }
  /*delay(del);
    vstep10 = ot.getBrinkTSP(TempIndoors);
    mqttClient.publish("brink/TempIndoors/get", String(vstep10/2).c_str());

    delay(del);
    vstep9 = ot.getBrinkTSP(TempAtmo) ;
    mqttClient.publish("brink/TempAtmo/get", String(vstep9/2).c_str());
  */
  delay(del);
  byPassStatus = ot.getBrinkTSP(I8);
  if (byPassStatus != byPassStatus_old) {
    mqttClient.publish("brink/I8/get", String(byPassStatus).c_str());
    byPassStatus_old = byPassStatus;
  }

}
/*

  // mqttClient.publish("brink/MinVol/get", String(ot.getBrinkTSP(MaxVol)).c_str());
*/

void scanner_TSP(int id) {

  unsigned long msg1 = ot.buildRequest(OpenThermRequestType::READ, OpenThermMessageID::VentTSPEntry, id << 8);
  unsigned long msg = ot.buildRequest(OpenThermRequestType::READ, OpenThermMessageID::VentTSPEntry, id << 8);
  unsigned long msg2 = ot.buildRequest(OpenThermRequestType::READ, OpenThermMessageID::VentTSPEntry, ( (unsigned int)id + 1) << 8);
  unsigned long response = ot.sendRequest(msg);
  unsigned long response1 = ot.sendRequest(msg1);
  unsigned long response2 = ot.sendRequest(msg2);
  //if (ot.isValidResponse(response))
  /*{
    unsigned int data = ot.getU8(response);
    //Serial.println("========");
    Serial.print("TSP ID: ");
    Serial.print(id);
    Serial.print(" ==> DATA: ");
    Serial.println(data);
  }
  else */
  //{
    response = ot.getU8(response);
    response2 = ot.getU8(response2);
    ot.getU8(response |= response2 << 8);
    //Serial.println("========");
    Serial.print("2TSP ID: ");
    Serial.print(id);
    Serial.print(" ==> DATA: ");
    Serial.println(response);
 // }
// else{
 //  Serial.println("INVALID");
 // }
}
void scan(int id) {

  unsigned long msg = ot.buildRequest(OpenThermRequestType::READ, (OpenThermMessageID)id, 0);
  unsigned long response = ot.sendRequest(msg);
  if (ot.isValidResponse(response)) {
    unsigned int data = ot.getUInt(response);
    Serial.print("ID: ");
    Serial.print(id);
    Serial.print(" ==> DATA: ");
    Serial.println(data);
  } else {
    //Serial.println("INVALID");
  }

}

void Monitor()
{
  Serial.print( String(cvol).c_str());
  Serial.print(", ");
  Serial.print(String(gvent * 4.5).c_str());
  Serial.print(", ");
  Serial.print(String(output).c_str());
  Serial.print(", ");
  Serial.print(String(input).c_str());
  Serial.print(", ");
  Serial.print(String(tIn).c_str());
  Serial.print(", ");
  Serial.print(String(tOut).c_str());
  Serial.print(", ");
  Serial.print(String(filter).c_str());
  Serial.print(", ");
  Serial.print(String(vstep1).c_str());
  Serial.print(", ");
  Serial.print(String(vstep2).c_str());
  Serial.print(", ");
  Serial.print(String(vstep3).c_str());
  Serial.print(", ");
  Serial.print(String(tU4 / 2).c_str());
  Serial.print(", ");
  Serial.print(String(tU5 / 2).c_str());
  Serial.print(", ");
  Serial.print(String(pbypass).c_str());
  Serial.print(", ");
  Serial.print(String(sbypass).c_str());
  Serial.print(", ");
  Serial.println(String(byPassStatus).c_str());
  Serial.print("==== ");
  Serial.println(String(cpod).c_str());
}
